import logging
import threading

from .Limits import LimitCollection, RawLimit


class HeaderBasedLimiter(object):
    def __init__(self, limit_header, count_header, friendly_name=None):
        self._limit_header = limit_header
        self._count_header = count_header
        self._friendly_name = friendly_name

        self._limits = {}
        self._limits_lock = threading.Lock()

    @property
    def friendly_name(self):
        return self._friendly_name

    def _get_limit_scope(self, region, endpoint_name, method_name):
        return ""

    def __get_limit(self, region, endpoint_name, method_name):
        scope = self._get_limit_scope(region, endpoint_name, method_name)

        if scope not in self._limits:
            with self._limits_lock:
                # double check this hasnt been added by another thread already
                if scope not in self._limits:
                    self._limits[scope] = LimitCollection()
                return self._limits[scope]

        return self._limits[scope]

    def wait_until(self, region, endpoint_name, method_name):
        scoped_limit = self.__get_limit(region, endpoint_name, method_name)

        return scoped_limit.wait_until()

    def update_limiter(self, region, endpoint_name, method_name, response):
        raw_limits = self._extract_headers(response)

        if raw_limits is None:
            return

        scoped_limit = self.__get_limit(region, endpoint_name, method_name)

        scoped_limit.update_limits(raw_limits)

    def _extract_headers(self, response):
        limits = HeaderBasedLimiter._extract_single_header(self._limit_header, response)
        counts = HeaderBasedLimiter._extract_single_header(self._count_header, response)

        if limits is None or counts is None:
            return None

        if len(limits) != len(counts):
            logging.warning(
                'header "%s" and "%s" have different sizes!',
                self._limit_header,
                self._count_header,
            )

        combined_limits = list(zip(counts, limits))

        for limit in combined_limits:
            if limit[0][1] != limit[1][1]:
                logging.warning(
                    'seems that limits for headers "%s" and "%s" did not match up correctly! '
                    + 'There may be issues in rate limiting. Headers were: "%s", "%s"'
                    + 'Limits from "%s" will be used.',
                    self._limit_header,
                    self._count_header,
                    response.headers.get(self._limit_header),
                    response.headers.get(self._count_header),
                    self._limit_header,
                )

        values = [
            RawLimit(count=limit[0][0], limit=limit[1][0], time=limit[0][1])
            for limit in combined_limits
        ]

        return values

    @staticmethod
    def _extract_single_header(header, response):
        values = response.headers.get(header)

        if values is None:
            return None

        values = values.split(",")

        values = [[int(val) for val in value.split(":")] for value in values]

        return values



from .RequestHandler import RequestHandler

from .DeprecationHandler import DeprecationHandler
from .JsonifyHandler import JsonifyHandler
from .TypeCorrectorHandler import TypeCorrectorHandler
from .ThrowOnErrorHandler import ApiError, ThrowOnErrorHandler
from .WaitingRateLimitHandler import WaitingRateLimitHandler

__all__ = [
    "RequestHandler",
    "DeprecationHandler",
    "JsonifyHandler",
    "ApiError",
    "ThrowOnErrorHandler",
    "TypeCorrectorHandler",
    "WaitingRateLimitHandler",
]



import logging

from riotwatcher.Handlers.RateLimit import RateLimitHandler


class WaitingRateLimitHandler(RateLimitHandler):
    def __init__(self):
        super(WaitingRateLimitHandler, self).__init__()
        logging.warning(
            "class WaitingRateLimitHandler is deprecated! "
            + "please use riotwatcher.Handlers.RateLimit.RateLimitHandler."
        )



import logging

from datetime import datetime

from . import RequestHandler


class DeprecationHandler(RequestHandler):
    def __init__(self):
        super(DeprecationHandler, self).__init__()
        self._warned = set()

    def after_request(self, region, endpoint_name, method_name, url, response):
        deprecation = response.headers.get("X-Riot-Deprecated")
        if deprecation is not None:
            try:
                deprecation = datetime.utcfromtimestamp(int(deprecation) // 1000)
            except (OSError, ValueError):
                # API returned unexected value in header, so just do nothing
                return response
            key = "{}.{}".format(endpoint_name, method_name)
            if key not in self._warned:
                # technically race condition here, but worst case is we double log a warning
                self._warned.add(key)
                logging.warning(
                    "Method %s has been deprecated by Riot! It will no longer work after %s",
                    key,
                    deprecation,
                )
        return response



from .Endpoint import RegionEndpoint


class SummonerV4Endpoint(RegionEndpoint):
    def __init__(self, url, **kwargs):
        nurl = "/summoner/v4/summoners" + url
        super(SummonerV4Endpoint, self).__init__(nurl, **kwargs)


class SummonerApiV4Urls(object):
    by_account = SummonerV4Endpoint("/by-account/{encrypted_account_id}")
    by_name = SummonerV4Endpoint("/by-name/{summoner_name}")
    by_puuid = SummonerV4Endpoint("/by-puuid/{encrypted_puuid}")
    by_id = SummonerV4Endpoint("/{encrypted_summoner_id}")



from .Endpoint import RegionEndpoint


class SpecatorV4Endpoint(RegionEndpoint):
    def __init__(self, url, **kwargs):
        nurl = "/spectator/v4" + url
        super(SpecatorV4Endpoint, self).__init__(nurl, **kwargs)


class SpectatorApiV4Urls(object):
    by_summoner = SpecatorV4Endpoint(
        "/active-games/by-summoner/{encrypted_summoner_id}"
    )
    featured_games = SpecatorV4Endpoint("/featured-games")



import datetime
import logging
import time

from .. import RequestHandler

from . import ApplicationRateLimiter, MethodRateLimiter, OopsRateLimiter


class RateLimitHandler(RequestHandler):
    def __init__(self):
        super(RateLimitHandler, self).__init__()

        self._limiters = (
            ApplicationRateLimiter(),
            MethodRateLimiter(),
            OopsRateLimiter(),
        )

    def preview_request(self, region, endpoint_name, method_name, url, query_params):
        """
        called before a request is processed.

        :param string region: the region of this request
        :param string endpoint_name: the name of the endpoint being requested
        :param string method_name: the name of the method being requested
        :param url: the URL that is being requested.
        :param query_params: dict: the parameters to the url that is being queried,
                                   e.g. ?key1=val&key2=val2
        """
        wait_until = max(
            [
                (
                    limiter.wait_until(region, endpoint_name, method_name),
                    limiter.friendly_name,
                )
                for limiter in self._limiters
            ],
            key=lambda lim_pair: lim_pair[0]
            if lim_pair[0]
            else datetime.datetime(datetime.MINYEAR, 1, 1),
        )

        if wait_until[0] is not None and wait_until[0] > datetime.datetime.now():
            to_wait = wait_until[0] - datetime.datetime.now()

            logging.info(
                "waiting for %s seconds due to %s limit...",
                to_wait.total_seconds(),
                wait_until[1],
            )
            time.sleep(to_wait.total_seconds())

    def after_request(self, region, endpoint_name, method_name, url, response):
        """
        Called after a response is received and before it is returned to the user.

        :param string region: the region of this request
        :param string endpoint_name: the name of the endpoint that was requested
        :param string method_name: the name of the method that was requested
        :param url: The url that was requested
        :param response: the response received. This is a response from the Requests library
        """
        for limiter in self._limiters:
            limiter.update_limiter(region, endpoint_name, method_name, response)

        return response



import re

from .UrlConfig import UrlConfig


class Endpoint(object):
    def __init__(self, url, **kwargs):
        self._url = url

        url_params = re.findall(r"{(\w*)}", self._url)

        if "" in url_params:
            raise ValueError("nameless format parameters not supported!")
        self._url_params = url_params
        self._query_params = [key for key in kwargs.keys() if key not in url_params]

    def __call__(self, **kwargs):
        for req_param in self._url_params:
            if req_param not in kwargs:
                raise ValueError('parameter "{}" is required!'.format(req_param))

        query_params = {
            key: value for key, value in kwargs.items() if key in self._query_params
        }

        return (self._url.format(**kwargs), query_params)


class RegionEndpoint(object):
    def __init__(self, url, **kwargs):
        self._url = "/lol" + url

    def __call__(self, **kwargs):
        final_url = UrlConfig.root_url + self._url

        endpoint = Endpoint(final_url, **kwargs)
        return endpoint(**kwargs)



from . import Endpoint


class DataDragonEndpoint(Endpoint):
    def __init__(self, url, **kwargs):
        nurl = "https://ddragon.leagueoflegends.com" + url
        super(DataDragonEndpoint, self).__init__(nurl, **kwargs)


class DDragonVersionLocaleEndpoint(DataDragonEndpoint):
    def __init__(self, url, **kwargs):
        nurl = "/cdn/{version}/data/{locale}" + url
        super(DDragonVersionLocaleEndpoint, self).__init__(nurl, **kwargs)


class DataDragonUrls(object):
    champions = DDragonVersionLocaleEndpoint("/champion.json")
    champions_full = DDragonVersionLocaleEndpoint("/championFull.json")
    items = DDragonVersionLocaleEndpoint("/item.json")
    languages = DDragonVersionLocaleEndpoint("/language.json")
    maps = DDragonVersionLocaleEndpoint("/map.json")
    masteries = DDragonVersionLocaleEndpoint("/mastery.json")
    profile_icons = DDragonVersionLocaleEndpoint("/profileicon.json")
    runes = DDragonVersionLocaleEndpoint("/rune.json")
    runes_reforged = DDragonVersionLocaleEndpoint("/runesReforged.json")
    summoner_spells = DDragonVersionLocaleEndpoint("/summoner.json")
    versions = DataDragonEndpoint("/realms/{region}.json")



from . import NamedEndpoint
from .urls import SummonerApiV4Urls


class SummonerApiV4(NamedEndpoint):
    """
    This class wraps the Summoner-v4 endpoint calls provided by the Riot API.

    See https://developer.riotgames.com/api-methods/#summoner-v4 for more detailed information
    """

    def __init__(self, base_api):
        """
        Initialize a new SummonerApiV4 which uses the provided base_api

        :param BaseApi base_api: the root API object to use for making all requests.
        """
        super(SummonerApiV4, self).__init__(base_api, self.__class__.__name__)

    def by_account(self, region, encrypted_account_id):
        """
        Get a summoner by account ID.

        :param string region:               The region to execute this request on
        :param string encrypted_account_id: The account ID.

        :returns: SummonerDTO: represents a summoner
        """
        url, query = SummonerApiV4Urls.by_account(
            region=region, encrypted_account_id=encrypted_account_id
        )
        return self._raw_request(self.by_account.__name__, region, url, query)

    def by_name(self, region, summoner_name):
        """
        Get a summoner by summoner name

        :param string region:           The region to execute this request on
        :param string summoner_name:    Summoner Name

        :returns: SummonerDTO: represents a summoner
        """
        url, query = SummonerApiV4Urls.by_name(
            region=region, summoner_name=summoner_name
        )
        return self._raw_request(self.by_name.__name__, region, url, query)

    def by_puuid(self, region, encrypted_puuid):
        """
        Get a summoner by PUUID.

        :param string region:           The region to execute this request on
        :param string encrypted_puuid:  PUUID

        :returns: SummonerDTO: represents a summoner
        """
        url, query = SummonerApiV4Urls.by_puuid(
            region=region, encrypted_puuid=encrypted_puuid
        )
        return self._raw_request(self.by_puuid.__name__, region, url, query)

    def by_id(self, region, encrypted_summoner_id):
        """
        Get a summoner by summoner ID.

        :param string region:                   The region to execute this request on
        :param string encrypted_summoner_id:    Summoner ID

        :returns: SummonerDTO: represents a summoner
        """
        url, query = SummonerApiV4Urls.by_id(
            region=region, encrypted_summoner_id=encrypted_summoner_id
        )
        return self._raw_request(self.by_id.__name__, region, url, query)



import requests

from .RequestHandler import RequestHandler

ApiError = requests.HTTPError


class ThrowOnErrorHandler(RequestHandler):
    def after_request(self, region, endpoint_name, method_name, url, response):
        try:
            response.raise_for_status()
        except requests.HTTPError as err:
            raise ApiError(*err.args, request=err.request, response=err.response)



from . import HeaderBasedLimiter


class ApplicationRateLimiter(HeaderBasedLimiter):
    def __init__(self):
        super(ApplicationRateLimiter, self).__init__(
            "X-App-Rate-Limit", "X-App-Rate-Limit-Count", "Application"
        )

    def _get_limit_scope(self, region, endpoint_name, method_name):
        return "{}".format(region)



from . import HeaderBasedLimiter


class MethodRateLimiter(HeaderBasedLimiter):
    def __init__(self):
        super(MethodRateLimiter, self).__init__(
            "X-Method-Rate-Limit", "X-Method-Rate-Limit-Count", "Method"
        )

    def _get_limit_scope(self, region, endpoint_name, method_name):
        return "{}/{}/{}".format(region, endpoint_name, method_name)



from .Endpoint import RegionEndpoint


class LeagueV4Endpoint(RegionEndpoint):
    def __init__(self, url, **kwargs):
        nurl = "/league/v4" + url
        super(LeagueV4Endpoint, self).__init__(nurl, **kwargs)


class LeagueApiV4Urls(object):
    challenger_by_queue = LeagueV4Endpoint("/challengerleagues/by-queue/{queue}")
    grandmaster_by_queue = LeagueV4Endpoint("/grandmasterleagues/by-queue/{queue}")
    by_id = LeagueV4Endpoint("/leagues/{league_id}")
    master_by_queue = LeagueV4Endpoint("/masterleagues/by-queue/{queue}")
    by_summoner = LeagueV4Endpoint("/entries/by-summoner/{encrypted_summoner_id}")
    entries = LeagueV4Endpoint("/entries/{queue}/{tier}/{division}", page=int)

    # deprecated
    positions_by_summoner = LeagueV4Endpoint(
        "/positions/by-summoner/{encrypted_summoner_id}"
    )



from . import NamedEndpoint
from .urls import SpectatorApiV4Urls


class SpectatorApiV4(NamedEndpoint):
    """
    This class wraps the Spectator-v4 endpoint calls provided by the Riot API.

    See https://developer.riotgames.com/api-methods/#spectator-v4 for more detailed information
    """

    def __init__(self, base_api):
        """
        Initialize a new SpectatorApiV3 which uses the provided base_api

        :param BaseApi base_api: the root API object to use for making all requests.
        """
        super(SpectatorApiV4, self).__init__(base_api, self.__class__.__name__)

    def by_summoner(self, region, encrypted_summoner_id):
        """
        Get current game information for the given summoner ID

        :param string region:                   The region to execute this request on
        :param string encrypted_summoner_id:    The ID of the summoner.

        :returns: CurrentGameInfo
        """
        url, query = SpectatorApiV4Urls.by_summoner(
            region=region, encrypted_summoner_id=encrypted_summoner_id
        )
        return self._raw_request(self.by_summoner.__name__, region, url, query)

    def featured_games(self, region):
        """
        Get list of featured games.

        :param string region: The region to execute this request on

        :returns: FeaturedGames
        """
        url, query = SpectatorApiV4Urls.featured_games(region=region)
        return self._raw_request(self.featured_games.__name__, region, url, query)



from . import RequestHandler


class JsonifyHandler(RequestHandler):
    def after_request(self, region, endpoint_name, method_name, url, response):
        return response.json()

    def after_static_request(self, url, response):
        return response.json()



from setuptools import setup, find_packages
import os.path

import sys

__version__ = "2.7.0"

descr_file = os.path.join(os.path.dirname(__file__), "README.rst")

dev_requirements = ["coverage", "pre-commit", "pytest", "pytest-cov", "tox"]

if sys.version_info < (3, 0):
    dev_requirements.append("mock")

setup(
    name="riotwatcher",
    version=__version__,
    packages=find_packages("src", exclude=["test"]),
    package_dir={"": "src"},
    description="RiotWatcher is a thin wrapper on top of the Riot Games API for League of Legends.",
    long_description=open(descr_file).read(),
    author="AG Stephan",
    url="https://github.com/pseudonym117/Riot-Watcher",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Text Processing",
        "Topic :: Games/Entertainment :: Real Time Strategy",
        "Topic :: Games/Entertainment :: Role-Playing",
    ],
    license="MIT",
    install_requires=["requests"],
    extras_require={"dev": dev_requirements},
)



from ._apis import BaseApi, DataDragonApi, ChampionApiV3, LolStatusApiV3
from ._apis import (
    ChampionMasteryApiV4,
    LeagueApiV4,
    MatchApiV4,
    SpectatorApiV4,
    SummonerApiV4,
    ThirdPartyCodeApiV4,
)
from .Handlers import (
    DeprecationHandler,
    JsonifyHandler,
    ThrowOnErrorHandler,
    TypeCorrectorHandler,
)

from .Handlers.RateLimit import RateLimitHandler

from ._apis.urls import UrlConfig


class RiotWatcher(object):
    """
    RiotWatcher class is intended to be the main interaction point with the RiotAPI.
    """

    def __init__(
        self, api_key=None, custom_handler_chain=None, timeout=None, kernel_url=None
    ):
        """
        Initialize a new instance of the RiotWatcher class.

        :param string api_key: the API key to use for this instance
        :param List[RequestHandler] custom_handler_chain:
                    RequestHandler chain to pass to the created BaseApi object.
                    This chain is called in order before any calls to the API, and called in
                    reverse order after any calls to the API.
                    If preview_request returns data, the rest of the call short circuits,
                    preventing any call to the real API and calling any handlers that have already
                    been run in reverse order.
                    This should allow for dynamic tiered caching of data.
                    If after_request returns data, that is the data that is fed to the next handler
                    in the chain.
                    Default chain is:
                        [
                            JsonifyHandler,
                            ThrowOnErrorHandler,
                            TypeCorrector,
                            RateLimitHandler,
                            DeprecationHandler
                        ]
        :param int timeout: Time to wait for a response before timing out a connection to
                            the Riot API
        :param string kernel_url: URL for the kernel instance to connect to, instead of the API.
                                  See https://github.com/meraki-analytics/kernel for details.
        """
        if not kernel_url and not api_key:
            raise ValueError("Either api_key or kernel_url must be set!")

        if custom_handler_chain is None:
            if kernel_url:
                custom_handler_chain = [
                    JsonifyHandler(),
                    ThrowOnErrorHandler(),
                    TypeCorrectorHandler(),
                    DeprecationHandler(),
                ]
            else:
                custom_handler_chain = [
                    JsonifyHandler(),
                    ThrowOnErrorHandler(),
                    TypeCorrectorHandler(),
                    RateLimitHandler(),
                    DeprecationHandler(),
                ]

        if kernel_url:
            UrlConfig.root_url = kernel_url
        else:
            UrlConfig.root_url = "https://{region}.api.riotgames.com"

        self._base_api = BaseApi(api_key, custom_handler_chain, timeout=timeout)

        self._champion = ChampionApiV3(self._base_api)
        self._lol_status = LolStatusApiV3(self._base_api)
        self._data_dragon = DataDragonApi(self._base_api)
        self._champion_mastery = ChampionMasteryApiV4(self._base_api)
        self._league = LeagueApiV4(self._base_api)
        self._match = MatchApiV4(self._base_api)
        self._spectator = SpectatorApiV4(self._base_api)
        self._summoner = SummonerApiV4(self._base_api)
        self._third_party_code = ThirdPartyCodeApiV4(self._base_api)
        # todo: tournament-stub
        # todo: tournament

    @property
    def champion_mastery(self):
        """
        Interface to the ChampionMastery Endpoint

        :rtype: ChampionMasteryApiV4
        """
        return self._champion_mastery

    @property
    def champion(self):
        """
        Interface to the Champion Endpoint

        :rtype: ChampionApiV3
        """
        return self._champion

    @property
    def league(self):
        """
        Interface to the League Endpoint

        :rtype: LeagueApiV4
        """
        return self._league

    @property
    def lol_status(self):
        """
        Interface to the LoLStatus Endpoint

        :rtype: LolStatusApiV3
        """
        return self._lol_status

    @property
    def match(self):
        """
        Interface to the Match Endpoint

        :rtype: MatchApiV4
        """
        return self._match

    @property
    def spectator(self):
        """
        Interface to the Spectator Endpoint

        :rtype: SpectatorApiV4
        """
        return self._spectator

    @property
    def data_dragon(self):
        """
        Interface to the DataDragon Endpoint

        :rtype: DataDragonApi
        """
        return self._data_dragon

    @property
    def summoner(self):
        """
        Interface to the Summoner Endpoint

        :rtype: SummonerApiV4
        """
        return self._summoner

    @property
    def third_party_code(self):
        """
        Interface to the Third Party Code Endpoint

        :rtype: ThirdPartyCodeApiV4
        """
        return self._third_party_code



from .BaseApi import BaseApi
from .NamedEndpoint import NamedEndpoint
from .ChampionApiV3 import ChampionApiV3
from .LolStatusApiV3 import LolStatusApiV3
from .DataDragonApi import DataDragonApi

from .ChampionMasteryApiV4 import ChampionMasteryApiV4
from .LeagueApiV4 import LeagueApiV4
from .MatchApiV4 import MatchApiV4
from .SpectatorApiV4 import SpectatorApiV4
from .SummonerApiV4 import SummonerApiV4
from .ThirdPartyCodeApiV4 import ThirdPartyCodeApiV4

__all__ = [
    "BaseApi",
    "NamedEndpoint",
    "ChampionApiV3",
    "LolStatusApiV3",
    "DataDragonApi",
    "ChampionMasteryApiV4",
    "LeagueApiV4",
    "MatchApiV4",
    "SpectatorApiV4",
    "SummonerApiV4",
    "ThirdPartyCodeApiV4",
]
